@@ -33,13 +33,14 @@
 char        *user,              /* user to run as */
             *group,             /* group to run as */
             *root_jail,         /* directory to chroot to */
-            *pid_name;          /* file to record pid in */
+            *pid_name,          /* file to record pid in */
+            *ctrl_name;         /* control socket name */
 
 int         alive_to,           /* check interval for resurrection */
             daemonize,          /* run as daemon */
             log_facility,       /* log facility to use */
-            log_level,          /* logging mode - 0, 1, 2 */
             print_log,          /* print log messages to stdout/stderr */
+            grace,              /* grace period before shutdown */
             control_sock;       /* control socket */
 
 SERVICE     *services;          /* global services (if any) */
@@ -50,10 +51,11 @@ regex_t HEADER,             /* Allowed header */
         CHUNK_HEAD,         /* chunk header line */
         RESP_SKIP,          /* responses for which we skip response */
         RESP_IGN,           /* responses for which we ignore content */
-        /* RESP_REDIR,         /* responses for which we rewrite Location */
         LOCATION,           /* the host we are redirected to */
         AUTHORIZATION;      /* the Authorisation header */
 
+static int  shut_down = 0;
+
 #ifndef  SOL_TCP
 /* for systems without the definition */
 int     SOL_TCP;
@@ -105,17 +107,42 @@ l_id(void)
 }
 
 /*
- * handle SIGTERM - exit
+ * handle SIGTERM/SIGQUIT - exit
  */
 static RETSIGTYPE
 h_term(const int sig)
 {
     logmsg(LOG_NOTICE, "received signal %d - exiting...", sig);
     if(son > 0)
-        kill(son, SIGTERM);
+        kill(son, sig);
+    if(ctrl_name != NULL)
+        (void)unlink(ctrl_name);
     exit(0);
 }
 
+/*
+ * handle SIGHUP/SIGINT - exit after grace period
+ */
+static RETSIGTYPE
+h_shut(const int sig)
+{
+    int         status;
+    LISTENER    *lstn;
+
+    logmsg(LOG_NOTICE, "received signal %d - shutting down...", sig);
+    if(son > 0) {
+        for(lstn = listeners; lstn; lstn = lstn->next)
+            close(lstn->sock);
+        kill(son, sig);
+        while(wait(&status) != son)
+            logmsg(LOG_ERR, "MONITOR: bad wait (%s)", strerror(errno));
+        if(ctrl_name != NULL)
+            (void)unlink(ctrl_name);
+        exit(0);
+    } else
+        shut_down = 1;
+}
+
 /*
  * Pound: the reverse-proxy/load-balancer
  *
@@ -143,13 +170,12 @@ main(const int argc, char **argv)
     print_log = 0;
     (void)umask(077);
     control_sock = -1;
-#ifndef  NO_SYSLOG
-    openlog("pound", LOG_CONS, LOG_DAEMON);
-#endif
+    log_facility = -1;
     logmsg(LOG_NOTICE, "starting...");
 
+    signal(SIGHUP, h_shut);
+    signal(SIGINT, h_shut);
     signal(SIGTERM, h_term);
-    signal(SIGINT, h_term);
     signal(SIGQUIT, h_term);
     signal(SIGPIPE, SIG_IGN);
 
@@ -162,15 +188,13 @@ main(const int argc, char **argv)
     l_init();
     CRYPTO_set_id_callback(l_id);
     CRYPTO_set_locking_callback(l_lock);
-    init_host_mut();
-    init_RSAgen();
+    init_timer();
 
     /* prepare regular expressions */
     if(regcomp(&HEADER, "^([a-z0-9!#$%&'*+.^_`|~-]+):[ \t]*(.*)[ \t]*$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
     || regcomp(&CHUNK_HEAD, "^([0-9a-f]+).*$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
     || regcomp(&RESP_SKIP, "^HTTP/1.1 100.*$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
     || regcomp(&RESP_IGN, "^HTTP/1.[01] (10[1-9]|1[1-9][0-9]|204|30[456]).*$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
-    /* || regcomp(&RESP_REDIR, "^HTTP/1.[01] 30[1237].*$", REG_ICASE | REG_NEWLINE | REG_EXTENDED) */
     || regcomp(&LOCATION, "(http|https)://([^/]+)(.*)", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
     || regcomp(&AUTHORIZATION, "Authorization:[ \t]*Basic[ \t]*([^ \t]*)[ \t]*", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
     ) {
@@ -190,6 +214,26 @@ main(const int argc, char **argv)
     /* read config */
     config_parse(argc, argv);
 
+    if(log_facility != -1)
+        openlog("pound", LOG_CONS | LOG_NDELAY, LOG_DAEMON);
+    if(ctrl_name != NULL) {
+        struct sockaddr_un  ctrl;
+
+        memset(&ctrl, 0, sizeof(ctrl));
+        ctrl.sun_family = AF_UNIX;
+        strncpy(ctrl.sun_path, ctrl_name, sizeof(ctrl.sun_path) - 1);
+        (void)unlink(ctrl.sun_path);
+        if((control_sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
+            logmsg(LOG_ERR, "Control \"%s\" create: %s", ctrl.sun_path, strerror(errno));
+            exit(1);
+        }
+        if(bind(control_sock, (struct sockaddr *)&ctrl, (socklen_t)sizeof(ctrl)) < 0) {
+            logmsg(LOG_ERR, "Control \"%s\" bind: %s", ctrl.sun_path, strerror(errno));
+            exit(1);
+        }
+        listen(control_sock, 512);
+    }
+
     /* open HTTP listeners */
     for(lstn = listeners, n_listeners = 0; lstn; lstn = lstn->next, n_listeners++) {
         int opt;
@@ -249,11 +293,11 @@ main(const int argc, char **argv)
         /* daemonize - make ourselves a subprocess. */
         switch (fork()) {
             case 0:
-#ifndef NO_SYSLOG
-                close(0);
-                close(1);
-                close(2);
-#endif
+                if(log_facility != -1) {
+                    close(0);
+                    close(1);
+                    close(2);
+                }
                 break;
             case -1:
                 logmsg(LOG_ERR, "fork: %s - aborted", strerror(errno));
@@ -323,18 +367,12 @@ main(const int argc, char **argv)
                 exit(1);
             }
 #endif
-            /* start resurector */
-            if(pthread_create(&thr, &attr, thr_resurect, NULL)) {
+            /* start timer */
+            if(pthread_create(&thr, &attr, thr_timer, NULL)) {
                 logmsg(LOG_ERR, "create thr_resurect: %s - aborted", strerror(errno));
                 exit(1);
             }
 
-            /* start the RSA stuff */
-            if(pthread_create(&thr, &attr, thr_RSAgen, NULL)) {
-                logmsg(LOG_ERR, "create thr_RSAgen: %s - aborted", strerror(errno));
-                exit(1);
-            }
-
             /* start the controlling thread (if needed) */
             if(control_sock >= 0 && pthread_create(&thr, &attr, thr_control, NULL)) {
                 logmsg(LOG_ERR, "create thr_control: %s - aborted", strerror(errno));
@@ -346,11 +384,18 @@ main(const int argc, char **argv)
 
             /* and start working */
             for(;;) {
+                if(shut_down) {
+                    logmsg(LOG_NOTICE, "shutting down...");
+                    for(lstn = listeners; lstn; lstn = lstn->next)
+                        close(lstn->sock);
+                    sleep(grace);
+                    logmsg(LOG_NOTICE, "grace period expired - exiting...");
+                    if(ctrl_name != NULL)
+                        (void)unlink(ctrl_name);
+                    exit(0);
+                }
                 for(lstn = listeners, i = 0; i < n_listeners; lstn = lstn->next, i++) {
-                    if(lstn->disabled)
-                        polls[i].events = 0;
-                    else
-                        polls[i].events = POLLIN | POLLPRI;
+                    polls[i].events = POLLIN | POLLPRI;
                     polls[i].revents = 0;
                 }
                 if(poll(polls, n_listeners, -1) < 0) {
@@ -367,6 +412,12 @@ main(const int argc, char **argv)
                                 /* may happen on FreeBSD, I am told */
                                 logmsg(LOG_WARNING, "HTTP connection prematurely closed by peer");
                                 close(clnt);
+                            } else if(lstn->disabled) {
+                                /*
+                                addr2str(tmp, MAXBUF - 1, &clnt_addr.sin_addr);
+                                logmsg(LOG_WARNING, "HTTP disabled listener from %s", tmp);
+                                */
+                                close(clnt);
                             } else {
                                 thr_arg *arg;
 
